/*
=========================================================================
	CREATURE.CPP
    Source code to the Creature class (PC, NPC, and Monster entities).
	Copyright (c) 2004-2007 by Daniel R. Collins. All rights reserved.
  May be distributed under the terms of GPL version 2 (see Copying.txt).
=========================================================================
*/

#include <stdafx.h>

#include "creature.h"
#include <assert.h>


/*
--------------------------------------------------------------------------
  Tokens
--------------------------------------------------------------------------
*/

// Statistics 

static const struct statTokenStruct { 
  int stat; 
  char *hdrAbbr; 
  char *hdrFull; 
} statToken[] = {
  {Creature::CR,      " CR ",     "Challenge Rating"},
  {Creature::AC,      " AC ",     "Armor Class"},
  {Creature::HP,      " hp ",     "Hit Points"},
  {Creature::SR,      " SR ",     "Spell Resistance"},
  {Creature::AL,      " AL ",     "Alignment"},
  {Creature::init,    " Init ",   "Initiative"},
  {Creature::speed,   " Spd ",    "Speed"},
  {Creature::size,    " Sz ",     "Size"}
};

static const int statTokenNum = 
  sizeof(statToken)/sizeof(statTokenStruct);


// Abilities

static const struct abilityTokenStruct { 
  int ability; 
  char *hdrAbbr; 
  char *hdrFull; 
} abilityToken[] = {
  {Creature::Str,     " Str ",  "Strength"},
  {Creature::Dex,     " Dex ",  "Dexterity"},
  {Creature::Con,     " Con ",  "Constitution"},
  {Creature::Int,     " Int ",  "Intelligence"},
  {Creature::Wis,     " Wis ",  "Wisdom"},
  {Creature::Cha,     " Cha ",  "Charisma"}
};

static const int abilityTokenNum = 
  sizeof(abilityToken)/sizeof(abilityTokenStruct);


// Saves

static const struct saveTokenStruct { 
  int save; 
  char *hdrAbbr; 
  char *hdrFull; 
} saveToken[] = {
  {Creature::Fort,  " Fort ",   "Fortitude"},
  {Creature::Ref,   " Ref ",    "Reflex"},
  {Creature::Fort,  " Will ",   "Will"}
};

static const int saveTokenNum = 
  sizeof(saveToken)/sizeof(saveTokenStruct);


// Alignments

static const struct alignTokenStruct { 
  char *abbrAln; 
  char *fullAln;
} alignToken[] = {
	{"LG", "lawful good"},
	{"LN", "lawful neutral"},
	{"LE", "lawful evil"},
	{"CG", "chaotic good"},
	{"CN", "chaotic neutral"},
	{"CE", "chaotic evil"},
	{"NG", "neutral good"},
	{"NE", "neutral evil"},
	{"L" , "lawful"},
	{"C",  "chaotic"},
	{"G",  "good"},
	{"E",  "evil"},
	{"N",  "neutral"},
	{"",   ""}
};

static const int alignTokenNum = 
  sizeof(alignToken)/sizeof(alignTokenStruct);


/*
--------------------------------------------------------------------------
  Constructors
--------------------------------------------------------------------------
*/

Creature::Creature (const char *data) {
  assert(data);
  const char *p;

  // Name goes up to colon
  if (p = strchr(data, ':')) {
    name.assign(data, p-data);
  }

  // Find main stats
  for (int i = 0; i < statTokenNum; i++) {
    statTokenStruct s = statToken[i];
    stat[s.stat] = 0;
    if (p = strstr(data, s.hdrAbbr)) {
      stat[s.stat] = atoi(p + strlen(s.hdrAbbr));
    }
  }

  // Hit Dice
  if (p = strstr(data, " HD ")) {
    HitDice.read(p+4);
  }

  // Attacks
  if (p = strstr(data, " Atk ")) {
    p += 5;
    readAttacks(p);
  }

  // Alignment
  stat[AL] = alignTokenNum - 1;
  if (p = strstr(data, " AL ")) {
    readAlignment(p+4);
  }

  // Initialize combat variables
  damage = 0; 
  subdual = 0;
}


Creature::~Creature (void) {
  meleeAttacks.clear();
  rangedAttacks.clear();
}


/*
--------------------------------------------------------------------------
  Accessors
--------------------------------------------------------------------------
*/

static void checkRange (int min, int val, int max) {
  assert(min <= val && val <= max);
}


int Creature::getStat (int statNum) {
  checkRange(0, statNum, numStats);
  return stat[statNum];
}


int Creature::getFace (int faceNum) {
  checkRange(0, faceNum, numFaces);
  return face[faceNum];
}


int Creature::getSave (int saveNum) {
  checkRange(0, saveNum, numSaves);
  return save[saveNum];
}


int Creature::getAbility (int abilityNum) {
  checkRange(0, abilityNum, numAbilities);
  return ability[abilityNum];
}


// Get number of weapons for category

int Creature::getNumWeapons (int category) {
  WeaponList *attacks = getWeaponList(category);
  return attacks->size();
}


// Get total number of attacks over all weapons

int Creature::getNumAttacks (int category) {
  WeaponList *attacks = getWeaponList(category);

  int numAttacks = 0;
  WeaponList::iterator i;
  for (i = attacks->begin(); i != attacks->end(); i++) {
    numAttacks += i->getNumAttacks();
  }
  return numAttacks;
}


// Access one weapon in category

Weapon *Creature::getWeapon (int category, int num) {
  assert(num < getNumAttacks(category));
  WeaponList *attacks = getWeaponList(category);

  int count = 0;
  WeaponList::iterator i;
  for (i = attacks->begin(); count < num; i++, count++);
  return &(*i);
}


/*
--------------------------------------------------------------------------
  Mutators
--------------------------------------------------------------------------
*/

// Roll Hit Dice for hit points

int Creature::rollHitPoints (void) {
  int roll = HitDice.rollMin1();
  if (roll > 0)
    stat[HP] = HitDice.rollMin1();
  // Else no HD: then use fixed hp number
  return stat[HP];
}


/*
--------------------------------------------------------------------------
  Helper Functions
--------------------------------------------------------------------------
*/


WeaponList *Creature::getWeaponList (int category) {
  assert(category == Weapon::melee || category == Weapon::ranged);
  return (category == Weapon::melee ? &meleeAttacks : &rangedAttacks);
}


void Creature::readAttacks (const char *data) {
  assert(data);
  const char *p = data;
  int category = Weapon::melee;

  while (*p) {
    readWeapon(p, category);

    // Skip past next comma or semicolon
    for (; *p && *p != ',' && *p != ';'; p++);
    if (*p == 0 || *p == ';') {
      break;
    }
    p += 2;
      
    // Switch to ranged at "or"
    if (strncmp(p, "or ", 3) == 0) {
      category++;
      p += 3;
    }
  }
}


void Creature::readWeapon (const char *data, int category) {
  assert(data);
  Weapon *weapon = new Weapon(data);
  WeaponList *attacks = getWeaponList(category);
  attacks->insert(attacks->end(), *weapon);
  delete weapon;
}


void Creature::readAlignment (const char *data) {
  assert(data);
  for (int i = 0; i < alignTokenNum; i++) {
    char *token = alignToken[i].abbrAln;
    if (strncmp(data, token, strlen(token)) == 0) {
      stat[AL] = i;
      break;
    }
  }
}


/*
--------------------------------------------------------------------------
  Combat Accessors
--------------------------------------------------------------------------
*/

int Creature::getHitPoints (void) {
  return stat[HP] - damage;
}


bool Creature::isDown (void) {
  return getHitPoints() <= 0;
}


bool Creature::isDead (void) {
  return getHitPoints() <= -10;

}


/*
--------------------------------------------------------------------------
  Combat Actions
--------------------------------------------------------------------------
*/

// Make a single attack, Nth in the category

int Creature::attack (int category, int num, Creature *target, int atkMod) {
  assert(category == Weapon::melee || category == Weapon::ranged);
  assert(num < getNumAttacks(category));
  WeaponList *attacks = getWeaponList(category);

  // Get the right weapon and attack number
  WeaponList::iterator i;
  for (i = attacks->begin(); i->getNumAttacks() <= num; i++) {
    num -= i->getNumAttacks();
  }

  // Make the attack
  int damage = i->oneAttack(num, target->getStat(AC), atkMod);
  target->takeDamage(damage);
  return damage;
}


int Creature::takeDamage (int amount) {
  damage += amount;  
  return amount;
}

