/*
=====================================================================
	OGCOMBAT.CPP
    Superstructure for simulating Open Gaming RPG fights.
	Copyright (c) 2004-2007 by Daniel R. Collins. All rights reserved.
  May be distributed under the terms of GPL version 2 (see Copying.txt).
=====================================================================
*/

#include <stdafx.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "ogcombat.h"
#include "dice.h"
#include "weapon.h"
#include "creature.h"


/*
---------------------------------------------------------------------
  Standard NPC Fighter Statistics
---------------------------------------------------------------------
*/

// Descriptions of stock NPC Fighters

static const char *fighterDesc[] = {
  "Ftr1: CR 1; AC 18; hp 12; Spd 20; Atk bastard sword +5 (1d10+2/19), or comp. longbow +2 (1d8/x3)", 
  "Ftr2: CR 2; AC 19; hp 19; Spd 20; Atk bastard sword +6 (1d10+2/19), or comp. longbow +4 (1d8+2/x3)", 
  "Ftr3: CR 3; AC 21; hp 27; Spd 20; Atk bastard sword +7 (1d10+2/19), or comp. longbow +5 (1d8+2/x3)", 
  "Ftr4: CR 4; AC 21; hp 34; Spd 20; Atk bastard sword +8 (1d10+5/19), or comp. longbow +6 (1d8+3/x3)", 
  "Ftr5: CR 5; AC 21; hp 42; Spd 20; Atk bastard sword +10 (1d10+5/19), or comp. longbow +7 (1d8+3/x3)", 
  "Ftr6: CR 6; AC 22; hp 49; Spd 20; Atk bastard sword +11/+6 (1d10+5/19), or comp. longbow +8/+3 (1d8+3/x3)", 
  "Ftr7: CR 7; AC 22; hp 57; Spd 20; Atk bastard sword +12/+7 (1d10+6/19), or comp. longbow +9/+4 (1d8+3/x3)", 
  "Ftr8: CR 8; AC 23; hp 64; Spd 20; Atk bastard sword +13/+8 (1d10+6/17), or comp. longbow +11/+6 (1d8+4/x3)", 
  "Ftr9: CR 9; AC 23; hp 72; Spd 20; Atk bastard sword +14/+9 (1d10+6/17), or comp. longbow +12/+7 (1d8+5/x3)", 
  "Ftr10: CR 10; AC 24; hp 79; Spd 20; Atk bastard sword +15/+10 (1d10+6/17), or comp. longbow +13/+8 (1d8+5/x3)", 
  "Ftr11: CR 11; AC 25; hp 87; Spd 20; Atk bastard sword +16/+11/+6 (1d10+6/17), or comp. longbow +14/+9/+4 (1d8+5/x3)", 
  "Ftr12: CR 12; AC 25; hp 94; Spd 20; Atk bastard sword +19/+14/+9 (1d10+8/17), or comp. longbow +16/+11/+6 (1d8+6/x3)", 
  "Ftr13: CR 13; AC 25; hp 102; Spd 20; Atk bastard sword +20/+15/+10 (1d10+8/17), or comp. longbow +17/+12/+7 (1d8+6/x3)", 
  "Ftr14: CR 14; AC 27; hp 109; Spd 20; Atk bastard sword +21/+16/+11 (1d10+8/17), or comp. longbow +18/+13/+8 (1d8+6/19/x3)", 
  "Ftr15: CR 15; AC 28; hp 117; Spd 20; Atk bastard sword +23/+18/+13 (1d10+9/17), or comp. longbow +19/+14/+9 (1d8+6/19/x3)", 
  "Ftr16: CR 16; AC 30; hp 124; Spd 20; Atk bastard sword +24/+19/+14/+9 (1d10+9/17), or comp. longbow +21/+16/+11/+6 (1d8+7/19/x3)", 
  "Ftr17: CR 17; AC 31; hp 132; Spd 20; Atk bastard sword +27/+22/+17/+12 (1d10+11/17), or comp. longbow +22/+17/+12/+7 (1d8+7/19/x3)", 
  "Ftr18: CR 18; AC 32; hp 139; Spd 20; Atk bastard sword +28/+23/+18/+13 (1d10+11/17), or comp. longbow +24/+19/+14/+9 (1d8+8/19/x3)", 
  "Ftr19: CR 19; AC 32; hp 166; Spd 20; Atk bastard sword +31/+26/+21/+16 (1d10+13/17), or comp. longbow +25/+20/+15/+10 (1d8+8/19/x3)", 
  "Ftr20: CR 20; AC 34; hp 175; Spd 20; Atk bastard sword +33/+28/+23/+18 (1d10+14/17), or comp. longbow +26/+21/+16/+11 (1d8+8/19/x3)"
};


// Return description of stock Nth-level NPC Fighter

const char *getFighterDesc(int level) {
  return (1 <= level && level <= 20 ? fighterDesc[level-1] : NULL);
}


/*
---------------------------------------------------------------------
  Constructors
---------------------------------------------------------------------
*/

// Standard constructor

OGCombat::OGCombat (const char *heroDesc, const char *enemyDesc, 
                    int enemies, int category)
{
  this->heroDesc = heroDesc;
  this->enemyDesc = enemyDesc;
  this->enemies = enemies;
  this->category = category;
}


// Stock NPC Fighter constructor

OGCombat::OGCombat (int ftrLevel, const char *enemyDesc, int enemies, int category)
{
  this->heroDesc = getFighterDesc(ftrLevel);
  this->enemyDesc = enemyDesc;
  this->enemies = enemies;
  this->category = category;
}


// Destructor

OGCombat::~OGCombat (void) {
}


/*
---------------------------------------------------------------------
  Functions
---------------------------------------------------------------------
*/

// Fight several combats: return percent hero won

int OGCombat::fight (int combats) {
  Dice::init();
  int heroWins = 0;
  for (int i = 0; i < combats; i++) {
    if (oneCombat()) {
      heroWins++;
    }
  }
  return 100 * heroWins / combats;
}


// One combat: returns whether hero won

int OGCombat::oneCombat () {

  // Construct creatures
  hero = new Creature(heroDesc);
  enemy = new Creature*[enemies];
  for (int i = 0; i < enemies; i++) {
    enemy[i] = new Creature(enemyDesc);
    enemy[i]->rollHitPoints();
  }

  // Run combat
  while (!hero->isDown() && enemiesKilled() < enemies) {
    enemyAttacks();
    heroAttacks();
  }
  int didHeroWin = (enemiesKilled() == enemies);

  // Delete creatures
  delete hero;
  for (i = 0; i < enemies; i++) {
    delete enemy[i];
  }
  delete enemy;
  return didHeroWin;
}


// Hero attack action

void OGCombat::heroAttacks () {
  if (!hero->isDown()) {
    int numDropped = 0;
    for (int atk = 0; atk < hero->getNumAttacks(category); atk++) {
      for (int target = 0; target < enemies; target++) {  // find target
        if (!enemy[target]->isDown() && numDropped < enemiesInRange()) {
          hero->attack(category, atk, enemy[target], 0);
          if (enemy[target]->isDown()) numDropped++;
          break;
        }
      }
    }
  }
}


// Enemy attack action

void OGCombat::enemyAttacks () {
  int numAttacked = 0;
  for (int foe = 0; foe < enemies; foe++) { // everyone in range attacks
    if (!enemy[foe]->isDown() && numAttacked < enemiesInRange()) {
      for (int atk = 0; atk < enemy[foe]->getNumAttacks(category); atk++) {
        enemy[foe]->attack(category, atk, hero, 0);
      }
      numAttacked++;
    }
  }
}


// Figure enemies in range of attacks

int OGCombat::enemiesInRange() {
  if (category == 0) {  // melee
    return enemiesInContact;
  } else {              // ranged
    return enemies;
  };
}


// Count enemies killed

int OGCombat::enemiesKilled () {
  int killed = 0;
  for (int i = 0; i < enemies; i++) {
    if (enemy[i]->isDown()) {
      killed++;
    }
  }
  return killed;
}

