/*
=========================================================================
	WEAPON.CPP
    Implementation of the Weapon class (attack routines).
	Copyright (c) 2004-2007 by Daniel R. Collins. All rights reserved.
  May be distributed under the terms of GPL version 2 (see Copying.txt).
=========================================================================
*/

#include <stdafx.h>

#include "weapon.h"
#include <assert.h>


/*
-------------------------------------------------------------------------
  Constructors
-------------------------------------------------------------------------
*/

Weapon::Weapon (const char *data) {
  assert(data);

  // Defaults
  natural = false;
  numAttacks = 1;
  critRange = 20;
  critMult = 2;

  // If starting with number, assume natural attack
  const char *p = data;
  if (isdigit(*p)) {
    natural = true;
    numAttacks = atoi(p);
    p = strchr(p, ' ');
    assert(p);
  }

  // Look for name of attack
  const char *n = p;
  while (*n && !isalpha(*n))
    n++;

  // Name goes up to the +/- in attack
  int nameLength = n - p + strcspn(n, "+-");
  name.assign(p, nameLength - 1);
  p += nameLength;

  // Primary attack bonus is next number
  primaryBonus = atoi(p);

  // Iterative attacks, count slashes up to paren
  while (*p && *p != '(') {
    if (!natural && *p == '/') {
      numAttacks++;
    }
    p++;
  }

  // Damage is after parenthesis
  assert(*p == '(');
  damage.read(p+1);

  // Criticals after a slash, end at comma or paren
  while (*p && *p != ',' && *p != ')') {
    if (*p == '/') {
      if (*(p+1) == 'x') {
        critMult = atoi(p+2);
      } 
      else {
        critRange = atoi(p+1);
      }
    }
    p++;
  }
}


/*
-------------------------------------------------------------------------
  Helper functions
-------------------------------------------------------------------------
*/

// Get attack bonus for Nth attack with this weapon

int Weapon::getAttackBonus (int num) {
  assert(0 <= num && num < numAttacks);

  if (natural) {
    // Assume natural attacks are all at same bonus.
    return primaryBonus;
  }
  else {
    // Assume other attacks go down by -5 per attack.
    return primaryBonus - 5 * num;
  }
}


/*
-------------------------------------------------------------------------
  Accessors
-------------------------------------------------------------------------
*/

int Weapon::getNumAttacks (void) {
  return numAttacks;
}


/*
-------------------------------------------------------------------------
  Attack functions
-------------------------------------------------------------------------
*/

// Make one attack action, return damage

int Weapon::oneAttack (int num, int AC, int atkMod, int crit) {
  assert(0 <= num && num < numAttacks);

  // Roll to hit
  int dieRoll = Dice::roll(20);
  int totalRoll = dieRoll + getAttackBonus(num) + atkMod;
  if ((totalRoll >= AC && dieRoll > 1) || dieRoll == 20) {
    int dam = 0;
    int times = (crit ? critMult-1 : 1);
    for (int i = 0; i < times; i++) {
      int damRoll = damage.roll();
      dam += (damRoll > 1 ? damRoll : 1);
    }

    // Check for critical threat
    if (!crit && dieRoll >= critRange) {
      dam += oneAttack(num, AC, atkMod, 1);
    }

    return dam;
  }
  return 0;
}


// Make a full attack action, return damage

int Weapon::fullAttack (int AC, int atkMod) {
  int dam = 0;
  for (int i = 0; i < numAttacks; i++) {
    dam += oneAttack(AC, atkMod, i);
  }
  return dam;
}

